#include <math.h>
#include <string.h>
#include "ukr.h"
#include "specialmath.h"

#ifndef DBL_MAX
#define DBL_MAX        1.7976931348623157E+308
#endif

Ukr_kp_dists_sparse ukr_get_kernel_sparse(UKR_Kernel kernel) {
   switch(kernel) {
      case UKR_KERNEL_QUARTIC:
         return ukr_kp_dists_quartic;
      case UKR_KERNEL_TRIWEIGHT:
         return ukr_kp_dists_triweight;
      default:
         return NULL;
   }
}

Ukr_kp_dists_full ukr_get_kernel_full(UKR_Kernel kernel) {
   switch(kernel) {
      case UKR_KERNEL_GAUSSIAN:
         return ukr_kp_dists_gaussian;
      default:
         return NULL;
   }
}

double ukr_kp_dists_gaussian(int M,const double *dist,double *k,double *p,int numExclude,const int *exclude) {
   int i=0,ne;
   double sum=0.0;
   
   for (ne=0;ne<numExclude;ne++) {
      for (;i<exclude[ne];i++) {
         sum+=k[i]= exp(-0.5*dist[i]);
      }         
      k[i++]=0;
   }
   for (;i<M;i++) {
      sum+=k[i]= exp(-0.5*dist[i]);
   }
   if (p!=NULL && p!=k) memcpy(p,k,M*sizeof(double));
   return sum;      
}
   
double ukr_kp_dists_quartic(int M,const double *dist,double *k,double *p,int *ir,int *nz,int numExclude,const int *exclude) {
   int i,ne;
   int n=0;
   double sum=0.0;
   
   if (p) {
      i=0;
      for (ne=0;ne<numExclude;ne++) {
         for (;i<exclude[ne];i++) {
            double h = 1-dist[i];         
            if (h>0) {
               p[n]=4.0*h;
               sum+=k[n]=h*h;
               ir[n]=i;
               n++;
            }         
         }
         i++;
      }
      for (;i<M;i++) {
         double h = 1-dist[i];      
         if (h>0) {
            p[n]=4.0*h;
            sum+=k[n]=h*h;
            ir[n]=i;
            n++;
         }
      }
   } else {
      i=0;
      for (ne=0;ne<numExclude;ne++) {
         for (;i<exclude[ne];i++) {
            double h = 1-dist[i];         
            if (h>0) {
               sum+=k[n]=h*h;
               ir[n]=i;
               n++;
            }         
         }
         i++;
      }
      for (;i<M;i++) {
         double h = 1-dist[i];         
         if (h>0) {
            sum+=k[n]=h*h;
            ir[n]=i;
            n++;
         }
      }
   }
   *nz=n;
   return sum;      
}

double ukr_kp_dists_triweight(int M,const double *dist,double *k,double *p,int *ir,int *nz,int numExclude,const int *exclude) {
   int i,ne;
   int n=0;
   double sum=0.0;
   
   if (p) {
      i=0;
      for (ne=0;ne<numExclude;ne++) {
         for (;i<exclude[ne];i++) {
            double h = 1-dist[i];         
            if (h>0) {
               double q = h*h;
               p[n]=6.0*q;
               sum+=k[n]=q*h;
               ir[n]=i;
               n++;
            }         
         }
         i++;
      }
      for (;i<M;i++) {
         double h = 1-dist[i];         
         if (h>0) {
            double q = h*h;
            p[n]=6.0*q;
            sum+=k[n]=q*h;
            ir[n]=i;
            n++;
         }
      }
   } else {
      i=0;
      for (ne=0;ne<numExclude;ne++) {
         for (;i<exclude[ne];i++) {
            double h = 1-dist[i];         
            if (h>0) {
               sum+=k[n]=h*h*h;
               ir[n]=i;
               n++;
            }         
         }
         i++;
      }
      for (;i<M;i++) {
         double h = 1-dist[i];         
         if (h>0) {
            sum+=k[n]=h*h*h;
            ir[n]=i;
            n++;
         }
      }
   }
   *nz=n;
   return sum;      
}

void ukr_kp_dists(const double *dist, UKR_Workspace *W, UKR_Model model, int numExclude, const int *exclude) {
   int i;
   switch(model.kernel) {
      case UKR_KERNEL_GAUSSIAN:
         W->sumK = ukr_kp_dists_gaussian(model.N, dist, W->K, W->P, numExclude,exclude);
         W->nz = model.N;
         for (i=0;i<model.N;i++) W->ir[i]=i;
         break;
      case UKR_KERNEL_QUARTIC:
         W->sumK = ukr_kp_dists_quartic(model.N, dist, W->K, W->P, W->ir, &W->nz, numExclude,exclude);
         break;
      case UKR_KERNEL_TRIWEIGHT:
         W->sumK = ukr_kp_dists_triweight(model.N, dist, W->K, W->P, W->ir, &W->nz, numExclude,exclude);
         break;
      default:
         W->sumK = 0;
         W->nz=0;
   }               
}

void ukr_aux_shiftX(double *Xx, int q, int nz, const int *ir) {
   int iz,m;
   
   for (iz=0;iz<nz;iz++) {
      for (m=0;m<q;m++) Xx[m+iz*q] = Xx[m+ir[iz]*q];
   }
}

int ukr_aux_b(UKR_Workspace *W, UKR_Model model) {
   int i,N;
   
   N=W->nz;
   
   if (W->sumK<=0) return 0;
   
   if (model.degree==0) {
      double h=1.0/W->sumK;
      for (i=0;i<N;i++) W->B[i]=W->K[i]*h;
      return 1;
   }
   W->C[0]=W->sumK;         
   
   switch(model.degree) {
      case 1:
         ukr_aux_l_C(W->C,W->Xx,W->K,model.q,N);
         if (invertSymmetricMatrix(W->C,1+model.q)==-1) return 0;
         ukr_aux_l_b(W->B,W->Xx,W->C,W->K,model.q,N);         
         break;
      case 2:
         ukr_aux_q_C(W->C,W->Xx,W->K,model.q,N);
         if (invertSymmetricMatrix(W->C,1+2*model.q)==-1) return 0;
         ukr_aux_q_b(W->B,W->Xx,W->C,W->K,model.q,N);
         break;
      case 3:
         ukr_aux_q2_C(W->C,W->Xx,W->K,N);
         if (invertSymmetricMatrix(W->C,6)==-1) return 0;
         ukr_aux_q2_b(W->B,W->Xx,W->C,W->K,N);
         break;
   }
   return 1;
}

void ukr_aux_l_b(double *b,const double *Xx,const double *invC,const double *K,int q,int N) {
   int i,m;
   for (i=0;i<N;++i) {
      double tmp=invC[0];
      for (m=0;m<q;m++) {
         double x = Xx[m+i*q];
         tmp+=x*invC[m+1];
      }
      b[i]=K[i]*tmp;
   }   
}

void ukr_aux_q_b(double *b,const double *Xx,const double *invC,const double *K,int q,int N) {
   int i,m;
   for (i=0;i<N;++i) {
      double tmp=invC[0];
      for (m=0;m<q;m++) {
         double x = Xx[m+i*q];
         tmp+=x*invC[m+1];
         tmp+=x*x*invC[m+q+1];
      }
      b[i]=K[i]*tmp;
   }
}

void ukr_aux_q2_b(double *b,const double *Xx,const double *invC,const double *K,int N) {
   int i,m;
   for (i=0;i<N;++i) {
      double x1 = Xx[0+i*2];
      double x2 = Xx[1+i*2];      
      b[i]=K[i]*(invC[0] + x1*invC[1] + x2*invC[2] + x1*x1*invC[3] + x2*x2*invC[4] + x1*x2*invC[5]);
   }
}

void multXiTMatrix_l(double *D,const double *Xx,const double *M,int q,int N,int d) {
   int i,j,m;
   for (j=0;j<d;j++) {
      double *Dj = D+j*N;
      const double *Mj = M+j*(1+q);
      for (i=0;i<N;++i) {
         const double *x = Xx + i*q;
         double Dij = Mj[0];
         for (m=0;m<q;m++) {
            Dij += x[m]*Mj[m+1];
         }
         Dj[i] = Dij;
      }
   }
}

void multXiTMatrix_q(double *D,const double *Xx,const double *M,int q,int N,int d) {
   int i,j,m;
   for (j=0;j<d;j++) {
      double *Dj = D+j*N;
      const double *Mj = M+j*(1+2*q);
      for (i=0;i<N;++i) {
         const double *x = Xx + i*q;
         double Dij = Mj[0];
         for (m=0;m<q;m++) {
            Dij += x[m]*(Mj[m+1] + x[m]*Mj[m+q+1]);
         }
         Dj[i] = Dij;
      }
   }
}

void multXiTMatrix_q2(double *D,const double *Xx,const double *M,int N,int d) {
   /* q=2 */
   int i,j;
   for (j=0;j<d;j++) {
      const double *Mj = M+j*6;
      for (i=0;i<N;++i) {
         double x1 = Xx[0+i*2];
         double x2 = Xx[1+i*2];
         D[i+j*N] = Mj[0] + Mj[1]*x1 + Mj[2]*x2 + Mj[3]*x1*x1 + Mj[4]*x2*x2 + Mj[5]*x1*x2;
      }
   }
}

void multXiMatrix_l(double *D, const double *Xx, const double *M, int q,int N,int d) {
   int j;
   
   for (j=0;j<d;j++) {
      double *Dj = D+j*(q+1);
      const double *Mj = M+j*N;
      
      Dj[0]=vectorSum1(Mj,N);
      multMatrixVector(Dj+1, Xx, Mj, q,N);
   }
}

void multXiMatrix_q(double *D, const double *Xx, const double *M, int q,int N,int d) {
   int m,j,i;
   
   for (j=0;j<d;j++) {
      double *Dj = D+j*(1+2*q);
      const double *Mj = M+j*N;
      Dj[0]=vectorSum1(Mj,N);
      for (m=0;m<q;m++) {
         double h = 0;
         double hq = 0;
         for (i=0;i<N;i++) {
            double x = Xx[m+i*q];
            double xm = x*Mj[i];
            h+=xm;
            hq+=x*xm;
         }
         Dj[1+m]=h;
         Dj[1+q+m]=hq;
      }
   }
}

void multXiMatrix_q2(double *D, const double *Xx, const double *M, int N,int d) {
   /* q=2 */
   int m,j,i;
   
   for (j=0;j<d;j++) {
      double *Dj = D+j*6;
      const double *Mj = M+j*N;
      Dj[0]=vectorSum1(Mj,N);
      for (m=1;m<6;m++) Dj[m]=0.0;
      for (i=0;i<N;i++) {
         double x1 = Xx[0+i*2];
         double x2 = Xx[1+i*2];
         double x1m = x1*Mj[i];
         double x2m = x2*Mj[i];
         Dj[1]+=x1m;
         Dj[2]+=x2m;
         Dj[3]+=x1*x1m;
         Dj[4]+=x2*x2m;
         Dj[5]+=x1*x2m;
      }
   }
}
  
int ukr_aux_bJ(double *J,UKR_Workspace *W,UKR_Model model) {
   int i,m,q1,q2,nz;
   double h;
   double *MuM=W->extra;
   nz = W->nz;
   
   q1=model.q+1;
   q2=model.q+q1;
      
   switch(model.degree) {
      case 0:
         h = 1.0/W->sumK;
         for (i=0;i<nz;i++) {
            W->P[i]*=h; 
            W->B[i]=W->K[i]*=h;
         }
         multMatrixVector(MuM, W->Xx, W->P,model.q,nz);
         for (m=0;m<model.q;m++) {
            for (i=0;i<nz;i++) {
               J[m*nz + i] = W->P[i]*W->Xx[m+i*model.q] - W->B[i]*MuM[m];
            }
         }
         break;
      case 1:
         /* Calculate U(:,1) as part of B */
         W->C[0]=W->sumK;         
         ukr_aux_l_C(W->C,W->Xx,W->K,model.q,nz);
         if (invertSymmetricMatrix(W->C,q1)==-1) return 0;
         
         multXiTMatrix_l(W->B, W->Xx, W->C, model.q, nz, 1);
         
         for (m=0;m<model.q;m++) {
            double *PXxU_KG = J + m*nz;    
            double Cm = W->C[1+m];     
            for (i=0;i<nz;i++) {
               PXxU_KG[i] = W->Xx[m+i*model.q]*W->P[i]*W->B[i] - W->K[i]*Cm;
            }
         }
         for (i=0;i<nz;i++) W->B[i]*=W->K[i];
         h=vectorSum1(W->B,nz);
         
         multXiMatrix_l(MuM,W->Xx,J,model.q,nz,model.q);
                  
         for (i=0;i<nz;i++) {
            double *H = W->Psi; 
            H[0] = W->B[i];
            for (m=1;m<q1;m++) {
               const double *G_m = W->C + m*q1;
               H[m] = W->K[i]*(G_m[0] + dotProduct(W->Xx + i*model.q, G_m + 1,model.q));
            }
            
            for (m=0;m<model.q;m++) {
               J[i + m*nz] += H[1+m]*h - dotProduct(H,MuM+m*q1,q1);
            }
         }
         break;
      case 2:
         /* Calculate U(:,1) as part of B */
         W->C[0]=W->sumK;         
         ukr_aux_q_C(W->C,W->Xx,W->K,model.q,nz);
         if (invertSymmetricMatrix(W->C,q2)==-1) return 0;
         
         multXiTMatrix_q(W->B, W->Xx, W->C, model.q, nz, 1);
         
         for (m=0;m<model.q;m++) {
            double *PXxU_KG = J + m*nz;    
            double Cm = W->C[1+m];     
            double Cmq = W->C[q1+m];
            for (i=0;i<nz;i++) {
               double x = W->Xx[m+i*model.q];
               PXxU_KG[i] = x*W->P[i]*W->B[i] - W->K[i]*Cm - 2*x*W->K[i]*Cmq;
            }
         }
         for (i=0;i<nz;i++) W->B[i]*=W->K[i];
         h=vectorSum1(W->B,nz);
         
         multXiMatrix_q(MuM,W->Xx,J,model.q,nz,model.q);
                  
         for (i=0;i<nz;i++) {
            double *H = W->Psi; 
            H[0] = W->B[i];
            for (m=1;m<q2;m++) {
               int mu;
               const double *G_m = W->C + m*q2;
               double help = G_m[0];
               for (mu=0;mu<model.q;mu++) {
                  double x = W->Xx[mu+i*model.q];
                  help += x*(G_m[1+mu] + x*G_m[q1+mu]);
               }
               H[m] = W->K[i]*help;
            }
            for (m=0;m<model.q;m++) {
               J[i + m*nz] += H[1+m]*h - dotProduct(H,MuM+m*q2,q2); 
            }
            
         }
         break;         
      case 3:
         /* Calculate U(:,1) as part of B */
         W->C[0]=W->sumK;         
         ukr_aux_q2_C(W->C,W->Xx,W->K,nz);
         if (invertSymmetricMatrix(W->C,6)==-1) return 0;
         
         multXiTMatrix_q2(W->B, W->Xx, W->C, nz, 1);
         
         /* PXxU_KG is held in J */
         for (i=0;i<nz;i++) {
            double x1 = W->Xx[0+i*2];
            double x2 = W->Xx[1+i*2];
            J[i+0*nz] = x1*W->P[i]*W->B[i] - W->K[i]*(W->C[1] + 2*(x1*W->C[3] + x2*W->C[5]));
            J[i+1*nz] = x2*W->P[i]*W->B[i] - W->K[i]*(W->C[2] + 2*(x2*W->C[4] + x1*W->C[5]));
         }
         for (i=0;i<nz;i++) W->B[i]*=W->K[i];
         h=vectorSum1(W->B,nz);
         
         multXiMatrix_q2(MuM,W->Xx,J,nz,2);
                  
         for (i=0;i<nz;i++) {
            double *H = W->Psi; 
            H[0] = W->B[i];
            for (m=1;m<6;m++) {
               const double *G_m = W->C + m*6;
               double x1 = W->Xx[0+i*2];
               double x2 = W->Xx[1+i*2];
               H[m] = W->K[i]*(G_m[0] + x1*(G_m[1] + x1*G_m[3] + x2*G_m[5]) + x2*(G_m[2] + x2*G_m[4]));
            }
            for (m=0;m<2;m++) {
               J[i + m*nz] += H[1+m]*h - dotProduct(H,MuM+m*6,6); 
            }
            
         }
         break;         
   }
   return 1;
}

void ukr_aux_l_C(double *C,const double *Xn,const double *K,int q,int N) {
   int i,nu,mu;
   int q1=q+1;
   
   multMatrixVector(C+1,Xn,K,q,N);
   for (nu=1;nu<q1;nu++) {      
      for (mu=0;mu<nu;mu++) {
         C[mu+nu*q1] = C[nu+mu*q1];
      }
      for (mu=nu;mu<q1;mu++) {
         double tmp=0.0;
         for (i=0;i<N;i++) {
            tmp+=K[i]*Xn[mu-1 + i*q]*Xn[nu-1 + i*q];               
         }
         C[mu+nu*q1] = tmp;
      }
   }
}

void ukr_aux_q_C(double *C,const double *Xn,const double *K,int q,int N) {
   int i,m,nu,mu;
   int q1=q+1;
   int q2=q+q1;
   
   for (m=0;m<q;m++) {
      double c1 = 0;
      double c2 = 0;
      for (i=0;i<N;++i) {
         double x = Xn[m+i*q];
         double k = K[i]*x;
         c1+=k;
         c2+=k*x;
      }
      C[1+m]=c1;
      C[q1+m]=c2;
   }
   for (nu=1;nu<q1;nu++) {
      for (mu=0;mu<nu;mu++) C[mu + nu*q2]=C[nu+mu*q2];
      for (mu=nu;mu<q1;mu++) {
         double c = 0;
         for (i=0;i<N;++i) {
            c+=K[i]*Xn[mu-1+i*q]*Xn[nu-1+i*q];
         }
         C[mu+nu*q2]=c;
      }
      for (mu=q1;mu<q2;mu++) {
         double c = 0;
         for (i=0;i<N;++i) {
            double x = Xn[mu-q1+i*q];
            c+=K[i]*x*x*Xn[nu-1+i*q];
         }
         C[mu+nu*q2]=c;
      }
   }   
   for (nu=q1;nu<q2;nu++) {
      for (mu=0;mu<nu;mu++) C[mu + nu*q2]=C[nu+mu*q2];
      for (mu=nu;mu<q2;mu++) {
         double c = 0;
         for (i=0;i<N;++i) {
            double x1 = Xn[mu-q1+i*q];
            double x2 = Xn[nu-q1+i*q];
            c+=K[i]*x1*x1*x2*x2;
         }
         C[mu+nu*q2]=c;
      }
   }         
}

void ukr_aux_q2_C(double *C,const double *Xn,const double *K,int N) {
   int i,m,nu,mu;
   /* C[0] is assumed to be already set */
   for (mu=1;mu<6*6;mu++) C[mu]=0;
   for (i=0;i<N;++i) {
      double x1 = Xn[i*2];
      double x2 = Xn[1+i*2];
      double kx1 = K[i]*x1;
      double kx2 = K[i]*x2;
      C[1+0*6]+=kx1;
      C[2+0*6]+=kx2;
      C[3+0*6]+=kx1*x1;
      C[4+0*6]+=kx2*x2;
      C[5+0*6]+=kx1*x2;
      C[1+1*6]+=kx1*x1;
      C[2+1*6]+=kx2*x1;
      C[3+1*6]+=kx1*x1*x1;
      C[4+1*6]+=kx2*x2*x1;
      C[5+1*6]+=kx2*x1*x1;
      C[2+2*6]+=kx2*x2;
      C[3+2*6]+=kx1*x1*x2;
      C[4+2*6]+=kx2*x2*x2;
      C[5+2*6]+=kx2*x1*x2;
      C[3+3*6]+=kx1*x1*x1*x1;
      C[4+3*6]+=kx2*x2*x1*x1;
      C[5+3*6]+=kx1*x2*x1*x1;
      C[4+4*6]+=kx2*x2*x2*x2;
      C[5+4*6]+=kx1*x2*x2*x2;
      C[5+5*6]+=kx1*x2*x1*x2;
   }
   
   for (nu=1;nu<6;nu++) {
      for (mu=0;mu<nu;mu++) C[mu + nu*6]=C[nu+mu*6];
   }
}

double ukr_aux_M(int col,UKR_Workspace *W,UKR_Model model) {
   int j;
   double funPart;
   double *M = W->M;
   const int *ir = W->ir;
   int nz = W->nz;
   
   if (model.G) {
      int i;
      const double *Gp = model.G + model.N*col;

      for (i=0;i<nz;i++) {
         M[i]=-Gp[ir[i]];
      }
      funPart=Gp[col];  
   
      for (j=0;j<nz;j++) {
         double Bj=W->B[j];
      
         Gp = model.G + model.N*ir[j]; 
         for (i=0;i<nz;i++) {
            M[i]+=Bj*Gp[ir[i]];
         }

         funPart-=Bj*Gp[col];
      }
      funPart+=dotProduct(M,W->B,nz);
      for (j=0;j<nz;j++) M[j]+=M[j]; /* *2 */
   } else {
      storeAlphaUtoV(W->B[0],model.Y+ir[0]*model.d,W->tmp_d,model.d);
      for (j=1;j<nz;j++) {
         addAlphaUtoV(W->B[j],model.Y+ir[j]*model.d,W->tmp_d,model.d);
      }
      
      funPart=ukr_loss_function(col, W->tmp_d, model.Yt + col*model.d, model);
      
      if (funPart>0) {
         for (j=0;j<nz;j++) {
            M[j]=dotProduct(W->tmp_d,model.Y+ir[j]*model.d,model.d);
         }
      } else {
         for (j=0;j<nz;j++) M[j]=0.0;
      }
   }
   return funPart;
}

double ukr_errgrad(double *GX, double *GXt,UKR_Model model,UKR_Workspace *W,int forceCV) {
   double fun=0.0;
   int i;
   int numExclude;
   const int *exclude;
   
   if (forceCV!=0) {
      numExclude=1;
      exclude=&i;  /* this is a trick to apply LOO-CV */
   } else if (model.MaskIs!=NULL) {
      numExclude=model.MaskJs[1];
      exclude=model.MaskIs;
   } else {
      numExclude=0;
      exclude=NULL;
   }
         
   for (i=0;i<model.Nt;i++) { 
      int nz;
      double funPart;
      
      shiftXcalcDists(W->K,W->Xx,model.X, model.Xt + i*model.q, model.q,model.N);
      ukr_kp_dists(W->K, W, model, numExclude,exclude);
      nz = W->nz;
      
      if (forceCV==0 && model.MaskIs!=NULL) {
         exclude+=numExclude;
         numExclude=model.MaskJs[i+2]-model.MaskJs[i+1];
      }
      
      if (nz==0) {
         if (model.Yt) {
            fun+=vectorSum2(model.Yt + i*model.d, model.d);
         } else {
            fun+=model.G[i*model.N + i];
         }
         continue;
      }      
      
      if (nz<model.N) ukr_aux_shiftX(W->Xx, model.q, nz, W->ir);
      
      ukr_aux_b(W,model);
      
      funPart = ukr_aux_M(i,W,model);
      fun+=funPart;
      
      if ((funPart>0) && (GX!=NULL)) {
         int m,j;
         
         ukr_aux_XQ_T(W,model.degree,model.q,model.Nt);
         
         for (j=0;j<nz;j++) {
            int off = model.q * W->ir[j];
            double *XQT = W->XQ_T + j*model.q;
            
            for (m=0;m<model.q;m++) {
               GX[m + off] -= XQT[m];
               GXt[m + model.q*i] += XQT[m];
            }
         }
      }
   }
   return fun/model.N;
}

double ukr_errgrad_scale(double *grad, const double *s,UKR_Model model,UKR_Workspace *W,int forceCV) {
   double fun=0.0;
   int i;
   int numExclude;
   const int *exclude;
   
   if (forceCV!=0) {
      numExclude=1;
      exclude=&i;  /* this is a trick to apply LOO-CV */
   } else if (model.MaskIs!=NULL) {
      numExclude=model.MaskJs[1];
      exclude=model.MaskIs;
   } else {
      numExclude=0;
      exclude=NULL;
   }
         
   for (i=0;i<model.Nt;i++) { 
      int nz;
      double funPart;
      
      shiftXcalcDists(W->K,W->Xx,model.X, model.Xt + i*model.q, model.q,model.N);
      ukr_kp_dists(W->K, W, model, numExclude,exclude);
      nz = W->nz;
      
      if (forceCV==0 && model.MaskIs!=NULL) {
         exclude+=numExclude;
         numExclude=model.MaskJs[i+2]-model.MaskJs[i+1];
      }
      
      if (nz==0) {
         if (model.Y) {
            fun+=vectorSum2(model.Yt + i*model.d, model.d);
         } else {
            fun+=model.G[i*model.N + i];
         }
         continue;
      }      
      
      if (nz<model.N) ukr_aux_shiftX(W->Xx, model.q, nz, W->ir);
      
      ukr_aux_b(W,model);
      
      funPart = ukr_aux_M(i,W,model);
      fun+=funPart;
      
      if ((funPart>0) && (grad!=NULL)) {
         int m,j;
         
         ukr_aux_XQ_T(W,model.degree,model.q,model.Nt);
         
         for (j=0;j<nz;j++) {
            double *XQT = W->XQ_T + j*model.q;
            double *Xx = W->Xx + j*model.q;
            
            for (m=0;m<model.q;m++) {
               grad[m] -= Xx[m]*XQT[m];
            }
         }
      }
   }
   if (grad!=NULL) {
      for (i=0;i<model.q;i++) {
         if (s[i]!=0.0) {
            grad[i]/=s[i];
         } else {
            grad[i]=0.0;
         }
      }
   }
   return fun/model.N;
}

void ukr_aux_XQ_T(UKR_Workspace *W, int degree, int q, int Nt) {
   int j,m;
   double UK;
   double Q;
   double PsiUK_MK;
   double PsiU;
   double U;
   double pre;
   
   double *Gamma = W->C;
   double *GammaPsi = W->GammaPsi;
   
   int nz = W->nz;
   
   double psi0 = 0.0;
   for (j=0;j<nz;++j) {
      psi0+=W->MK[j]=W->M[j]*W->K[j];
   }
   W->Psi[0] = psi0;   
   
   pre=1.0/Nt;
   
   switch (degree) {
      case 0:
         U = pre / W->sumK;
         PsiU = psi0/ W->sumK;
         for (j=0;j<nz;j++) {
            Q=(W->M[j]-PsiU)*W->P[j]*U;    
            for (m=0;m<q;m++) {
               W->XQ_T[m+j*q] = W->Xx[m+j*q] * Q;
            }
         }
         break;
      case 1:
         multMatrixVector(W->Psi+1,W->Xx,W->MK,q,nz);
         multMatrixVector(GammaPsi,Gamma,W->Psi,q+1,q+1);
         
         for (m=0;m<1+q;m++) {
            Gamma[m]*=pre;
         }
         
         for (j=0;j<nz;j++) {
            const double *x = W->Xx + j*q;
            
            PsiU = GammaPsi[0];
            for (m=0;m<q;m++) PsiU+=GammaPsi[1+m]*x[m];
            U = Gamma[0];
            for (m=0;m<q;m++) U+=Gamma[1+m]*x[m];
            UK= U*W->K[j];
            
            Q=(W->M[j] - PsiU)*W->P[j]*U;    
            PsiUK_MK = PsiU*W->K[j] - W->MK[j];        

            for (m=0;m<q;++m) {
               W->XQ_T[m+j*q] = x[m]*Q + GammaPsi[1+m]*UK + Gamma[1+m]*(PsiUK_MK);
            }
         }
         break;
      case 2:
         for (m=0;m<q;m++) {
            double Psi_m = 0;
            double Psi_mq = 0;
            for (j=0;j<nz;++j) {
               double x = W->Xx[m+j*q];
               double help = W->MK[j]*x;
               Psi_m+=help;
               Psi_mq+=help * x;
            }
            W->Psi[1+m]=Psi_m;
            W->Psi[q+1+m]=Psi_mq;
         } 
         multMatrixVector(GammaPsi,Gamma,W->Psi,2*q+1,2*q+1);
         for (m=0;m<1+2*q;m++) {
            Gamma[m]*=pre;
         }
         for (j=0;j<nz;j++) {
            const double *x = W->Xx + j*q;
            
            PsiU = GammaPsi[0];
            U = Gamma[0];            
            for (m=0;m<q;m++) {
               PsiU+=x[m]*(GammaPsi[1+m] + GammaPsi[q+1+m]*x[m]);
               U   +=x[m]*(Gamma[1+m] + Gamma[q+1+m]*x[m]);
            }
            UK= U*W->K[j];
            Q=(W->M[j] - PsiU)*W->P[j]*U;    
            PsiUK_MK = PsiU*W->K[j] - W->MK[j];        
            for (m=0;m<q;++m) {
               double Gm0 = Gamma[1+m] + 2*Gamma[q+1+m]*x[m];
               double Gma = GammaPsi[1+m] + 2*GammaPsi[q+1+m]*x[m];
               W->XQ_T[m+j*q] = Gma*UK + Gm0*(PsiUK_MK) + x[m]*Q;
            }
         }
         break;
      case 3:
         for (m=1;m<6;m++) W->Psi[m]=0.0;
         for (j=0;j<nz;++j) {
            double x1 = W->Xx[0+j*2];
            double x2 = W->Xx[1+j*2];
            double h1 = W->MK[j]*x1;
            double h2 = W->MK[j]*x2;
            W->Psi[1]+=h1;
            W->Psi[2]+=h2;
            W->Psi[3]+=h1*x1;
            W->Psi[4]+=h2*x2;
            W->Psi[5]+=h1*x2;
         }
         multMatrixVector(GammaPsi,Gamma,W->Psi,6,6);
         for (m=0;m<6;m++) {
            Gamma[m]*=pre;
         }
         for (j=0;j<nz;j++) {
            double x1,x2;
            
            x1 = W->Xx[0+j*2];
            x2 = W->Xx[1+j*2];
                        
            PsiU= GammaPsi[0]+ x1*(GammaPsi[1] + GammaPsi[3]*x1 + GammaPsi[5]*x2) 
                  + x2*(GammaPsi[2] + GammaPsi[4]*x2);
            U = Gamma[0] + x1*(Gamma[1] + Gamma[3]*x1 + Gamma[5]*x2)
                  + x2*(Gamma[2] + Gamma[4]*x2);
                   
            UK= U*W->K[j];
            
            Q=(W->M[j] - PsiU)*W->P[j]*U;    
            PsiUK_MK = PsiU*W->K[j] - W->MK[j];  
            {
               double G10 = Gamma[1] + 2*Gamma[3]*x1 + Gamma[5]*x2;
               double G1a = GammaPsi[1] + 2*GammaPsi[3]*x1 + GammaPsi[5]*x2;
               W->XQ_T[0+j*2] = G1a*UK + G10*PsiUK_MK + x1*Q;
            }
            {  
               double G20 = Gamma[2] + 2*Gamma[4]*x2 + Gamma[5]*x1;
               double G2a = GammaPsi[2] + 2*GammaPsi[4]*x2 + GammaPsi[5]*x1;
               W->XQ_T[1+j*2] = G2a*UK + G20*PsiUK_MK + x2*Q;
            }
         }
         break;
   }
}

double ukr_loss_function(int col, double *yr,const double *y,UKR_Model model) {
   double funPart = 0.0;
   double eps;
   double res;
   int j;
   
   switch(model.lossFunction) {

      case UKR_LOSS_L2:
      
         for (j=0;j<model.d;j++) {
            double h = yr[j]-y[j];
            yr[j] = h+h;  /* (|yr-y|^2)' = 2*(yr-y) */
            funPart+= h*h;
         }
         break;

      case UKR_LOSS_EPS:
      
         if (model.epsM>1) {
            const double *epss;
            if (model.epsN>1) {
               /* model.epsilon = dxN matrix */
               epss = model.epsilon + col*model.d;
            } else {
               /* model.epsilon = dx1 column vector */
               epss = model.epsilon;
            }
            for (j=0;j<model.d;j++) {
               double h = yr[j] - y[j];
               double ah = fabs(h) - epss[j];
               if (ah > 0) {
                  yr[j]=(h>0)?2*ah:-2*ah;
                  funPart += ah*ah;
               } else {
                  yr[j]=0;
               }
            }
         } else { 
            /* model.epsilon is scalar */
            eps = model.epsilon[0];
            for (j=0;j<model.d;j++) {
               double h = yr[j] - y[j];
               double ah = fabs(h) - eps;
               if (ah > 0) {
                  yr[j]=(h>0)?2*ah:-2*ah;
                  funPart += ah*ah;
               } else {
                  yr[j]=0;
               }
            }
         }
         break;
         
      case UKR_LOSS_EPS2:
         
         if (model.epsN>1) {
            /* model.epsilon = 1xN row vector */
            eps = model.epsilon[col];
         } else {
            /* model.epsilon is scalar */         
            eps = model.epsilon[0];
         }
         
         res = 0.0;

         for (j=0;j<model.d;j++) {
            double h = yr[j]-=y[j]; /* -= is ok */
            res+=h*h;
         }
         res = sqrt(res);
         if (res > eps) {
            funPart=(res-eps)*(res-eps);
            for (j=0;j<model.d;j++) {
               yr[j] *= 2.0*funPart/res;
            }
         } /* else funPart = 0, yr is ignored */
         break;
         
      case UKR_LOSS_HUBER:

         if (model.epsM>1) {
            /* model.epsilon = dx1 column vector */         
            for (j=0;j<model.d;j++) {
               double r = yr[j]-y[j];            
               eps = model.epsilon[j];
               if (r > eps) {
                  funPart+=r - 0.5*eps;
                  yr[j]=1;
               } else if (r < -eps) {
                  funPart+=-r + 0.5*eps;
                  yr[j]=-1;
               } else {
                  funPart+=r*r/(2*eps);
                  yr[j]=r/eps;
               }
            }
         } else {
            /* model.epsilon is scalar */                  
            eps = model.epsilon[0];         
            for (j=0;j<model.d;j++) {
               double r = yr[j]-y[j];
               if (r > eps) {
                  funPart+=r - 0.5*eps;
                  yr[j]=1;
               } else if (r < -eps) {
                  funPart+=-r + 0.5*eps;
                  yr[j]=-1;
               } else {
                  funPart+=r*r/(2*eps);
                  yr[j]=r/eps;
               }
            }
         }
         break;
         
      default: 
         return 0;
   }
   return funPart;
}

UKR_Kernel ukr_kernel_by_name(const char *name) {
   if (strcmp(name,"Gaussian")==0) return UKR_KERNEL_GAUSSIAN;
   if (strcmp(name,"Quartic")==0) return UKR_KERNEL_QUARTIC;
   if (strcmp(name,"Triweight")==0) return UKR_KERNEL_TRIWEIGHT;    
   return UKR_KERNEL_UNKNOWN;
}

UKR_Loss ukr_loss_by_name(const char *name) {
   if (strcmp(name,"L2")==0) return UKR_LOSS_L2;
   if (strcmp(name,"Eps")==0) return UKR_LOSS_EPS;
   if (strcmp(name,"Eps2")==0) return UKR_LOSS_EPS2;
   if (strcmp(name,"Huber")==0) return UKR_LOSS_HUBER;
   return UKR_LOSS_UNKNOWN;
}

UKR_DensPenalty ukr_dens_penalty(double *GX, double *GXt,UKR_Model model, UKR_Workspace *W) {
   UKR_DensPenalty res;
   int i;
   
   res.penalty=0.0;
   res.minDens=DBL_MAX;
   res.numViolate=0;
   
   for (i=0;i<model.Nt;i++) {
      int nz;
      double sumK;
      
      shiftXcalcDists(W->K,W->Xx,model.X, model.Xt + i*model.q, model.q,model.N);
      ukr_kp_dists(W->K, W, model, 0, NULL);
      
      nz = W->nz;
      sumK = W->sumK;
      
      if (sumK<res.minDens) res.minDens=sumK;
      sumK-=model.N*model.eta;

      if (sumK > 0) {
         res.penalty-=log(sumK/model.N);
         scaleVector(-1.0/sumK,W->P,nz);
         
         if (GX!=NULL) {
            int m,j;
            double *GXti = GXt + i*model.q;
            
            for (j=0;j<nz;j++) {
               int off = model.q * W->ir[j];
               const double *x = W->Xx + off;
               
               for (m=0;m<model.q;m++) {
                  double px = W->P[j]*x[m];
                  GX[m + off]-=px;
                  GXti[m]+=px;
               }
            }
         }
      } else {
         res.numViolate++;
      }
   }
   res.minDens/=model.N;
   return res;
}

double ukr_densvar_penalty(double *GX,double *GXt,UKR_Model model,UKR_Workspace *W) {
   int i;
   double penalty = 0.0;
   double sD=0.0,sD2=0.0;
   double md;
   
   /* Simple & stupid algorithm: First calculate mean density, 
   ** then run again for the gradient 
   ** Running time could be improved by storing intermediate results, costs memory
   */
   
   for (i=0;i<model.Nt;i++) { 
      calcSqDistances(W->K,model.X,model.Xt + i*model.q,model.q,model.N);
      ukr_kp_dists(W->K, W, model, 0, NULL);
      md = W->sumK/model.N;
      sD +=md;
      sD2+=md*md;
   }
   md = sD/model.Nt;
   
   penalty = 0.5*model.Nt*(sD2 - 2*sD*md + model.Nt*md*md);
   
   if (GX==NULL) return penalty;
   
   for (i=0;i<model.Nt;i++) { 
      int nz;
      double sumK;
      
      shiftXcalcDists(W->K,W->Xx,model.X, model.Xt + i*model.q, model.q,model.N);
      ukr_kp_dists(W->K, W, model, 0, NULL);
      
      nz = W->nz;
      
      if (nz>0) {
         int m,j;
         double fac = W->sumK/model.N - md;
         double *GXti = GXt + i*model.q;
         
         for (j=0;j<nz;j++) {
            int off = model.q * W->ir[j];
            const double *x = W->Xx + off;
            for (m=0;m<model.q;m++) {
               double px = fac*W->P[j]*x[m];
               GX[m + off]-=px;
               GXti[m]+=px;
            }
         }
      }
   }
   return penalty;
}
  
double ukr_f(double *f,const double *x,UKR_Model model, UKR_Workspace *W) {
   if (model.d==0) return -1;
   
   shiftXcalcDists(W->K,W->Xx,model.X,x,model.q,model.N);
   ukr_kp_dists(W->K,W,model,0,NULL);
   if (W->nz<model.N) ukr_aux_shiftX(W->Xx, model.q, W->nz, W->ir);
   if (ukr_aux_b(W,model)) {
      multMatrixVectorSp(f,model.Y,W->B,model.d,W->nz,W->ir);      
   } else {
      return -1;
   } 
   return W->sumK/model.N;
}

double ukr_fJ(double *f,double *J,const double *x,UKR_Model model, UKR_Workspace *W) {
   int i,m;
   double *Jb = W->XQ_T;
   const double *Yi;
   
   if (model.d==0) return -1;
         
   shiftXcalcDists(W->K,W->Xx,model.X,x,model.q,model.N);
   ukr_kp_dists(W->K,W,model,0,NULL);
   
   if (W->nz<model.N) ukr_aux_shiftX(W->Xx, model.q, W->nz, W->ir);
   if (!ukr_aux_bJ(Jb,W,model)) return -1.0;
   
   multMatrixVectorSp(f,model.Y,W->B,model.d,W->nz,W->ir);
   for (m=0;m<model.q;m++) {
      multMatrixVectorSp(J + m*model.d, model.Y, Jb + m*W->nz,model.d,W->nz,W->ir);
   }
   return W->sumK/model.N;
}

int ukr_aux_alloc_workspace(UKR_Model model, UKR_Workspace *W) {
   int n,nEx;

   W->ir=(int *) malloc(sizeof(int) * (model.N+model.q));
   if (W->ir==NULL) return 0;

   switch(model.degree) {
      case 0: 
         n=1; break;
      case 1: 
         n=model.q+1; break;
      case 2: 
         n=2*model.q+1; break;
      case 3: 
         n=6; break;
      default: 
         n=0;
   }
   
   nEx=model.N;
   if (n>nEx) nEx=n;
   if (model.q>nEx) nEx=model.q;
   if (model.d>nEx) nEx=model.d;      
   nEx*=model.q;

   W->B = malloc(sizeof(double) * (model.N*5 + model.d + n*n + 2*n + 2*model.q*model.N + nEx + model.q*(model.q+4)));
   if (W->B==NULL) {
      free(W->ir);
      W->ir=NULL;
      return 0;
   }
   W->nonbind = W->ir+model.N;
   W->K=W->B + model.N;
   W->P=W->K + model.N;
   W->M=W->P + model.N;
   W->MK=W->M + model.N;
   W->tmp_d = W->MK + model.N;
   W->C=W->tmp_d + model.d;
   W->Psi=W->C + n*n;
   W->GammaPsi=W->Psi + n;
   W->Xx=W->GammaPsi + n;
   W->XQ_T=W->Xx + model.q*model.N;
   W->extra=W->XQ_T + model.q*model.N;
   W->extraG=W->extra+nEx;
   return 1;
}

void ukr_aux_free_workspace(UKR_Workspace W) {
   if (W.B != NULL) free(W.B);
   if (W.ir!= NULL) free(W.ir);
}

double ukr_g_errgradhess(double *g,double *H,const double *x,UKR_Model model, UKR_Workspace *W, UKR_G_Struct gs, int kp_ready) {
	int i,j;
   int q=model.q;
   double *Jb = W->XQ_T;
   
   if (!kp_ready) {
      shiftXcalcDists(W->K,W->Xx,model.X,x,q,model.N);
      ukr_kp_dists(W->K,W,model,0,NULL);
   }
   if (W->nz<model.N) ukr_aux_shiftX(W->Xx, model.q, W->nz, W->ir);
   if (W->nz==0 || 0==ukr_aux_bJ(Jb,W,model)) {
      memset(g,0,sizeof(double)*q);
      memset(H,0,sizeof(double)*q*q);
      return (model.Y==NULL)?0.5*gs.Kyy:0.5*vectorSum2(gs.y,model.d);
   }
      
   if (model.Y!=NULL) {
      double *Jf = W->extra;
      
      multMatrixVectorSp(W->tmp_d,model.Y,W->B,model.d,W->nz,W->ir);
      for (i=0;i<model.q;i++) {
         multMatrixVectorSp(Jf + i*model.d, model.Y, Jb + i*W->nz,model.d,W->nz,W->ir);
      }
      
      for (i=0;i<model.d;i++) {
         W->tmp_d[i]-=gs.y[i];
      }
      
      if (model.mmm != NULL) {
         for (i=0;i<model.d;i++) {
            double mi = model.mmm[i];
            
            W->tmp_d[i]*=mi;
            for (j=0;j<model.q;j++) {
               Jf[i + j*model.d]*=mi;
            }
         }
      }
      
      multMatrixTVector(g,Jf,W->tmp_d,model.d,q);
      for (j=0;j<q;j++) {	
         for (i=0;i<j;i++) H[i+j*q]=H[j+i*q];
         for (i=j;i<q;i++) {
            H[i+j*q]=dotProduct(Jf+i*model.d,Jf+j*model.d,model.d);
         }
      }
      return 0.5*vectorSum2(W->tmp_d,model.d);
   } else {
      double *GJ = W->extra;
      int nz = W->nz;
      
      for (i=0;i<nz;i++) {
         double Gbi=0.0;
         const double *G_i = model.G+W->ir[i]*model.N;
         for (j=0;j<W->nz;j++) {
            /* Gbi+=G[ir[i] + ir[j]*N]*b[j]; */
            /* the same, but should be faster: */
            Gbi+=G_i[W->ir[j]]*W->B[j];               
         }
         W->M[i]=Gbi - 2*gs.KyY[W->ir[i]];
      }
      for (j=0;j<q;j++) {
         double *J_j = Jb + j*nz;
         for (i=0;i<nz;i++) {
            double GJij=0.0;
            const double *G_i = model.G + W->ir[i]*model.N;
            int k;
            for (k=0;k<nz;k++) {
               GJij+=G_i[W->ir[k]]*J_j[k];
            }
            GJ[i+j*nz]=GJij;
         }
      }

      for (i=0;i<q;i++) {
         g[i]=0.5 * (dotProduct(Jb+i*nz,W->M,nz) + dotProduct(GJ+i*nz,W->B,nz));
      }

      for (j=0;j<q;j++) {	
         for (i=0;i<j;i++) H[i+j*q]=H[j+i*q];
         for (i=j;i<q;i++) {
            H[i+j*q]=dotProduct(Jb+i*nz,GJ+j*nz,nz);
         }
      }

      return 0.5*(dotProduct(W->B,W->M,nz) + gs.Kyy);
   }
}

double ukr_g_congrad(double *g,const double *x,UKR_Model model, UKR_Workspace *W) {
   int i;
   double invN=-1.0/model.N;
   
   shiftXcalcDists(W->K,W->Xx,model.X,x,model.q,model.N);
   ukr_kp_dists(W->K,W,model,0,NULL);
      
   if (W->nz==0) {
      memset(g,0,sizeof(double)*model.q);
      return model.eta;
   }
     
   storeAlphaUtoV(invN*W->P[0],W->Xx + model.q*W->ir[0],g,model.q);
   for (i=1;i<W->nz;i++) {
      addAlphaUtoV(invN*W->P[i],W->Xx + model.q*W->ir[i],g,model.q);
   }
      
   return model.eta + W->sumK*invN;
}




#define KKT_SENS    0.99
#define CON_SENS    -1e-5

#define GRAD_TOL    1e-12
#define MAX_ITER    50
#define MIN_STEP    1e-12

double ukr_g_opt_unc(double *x, UKR_Model model, UKR_Workspace *W, UKR_G_Struct gs) {
   int iter;
   double oldf,f,ngf,pred,lambda;
   double *H,*gf,*dx;
   
   dx = W->extraG;
   gf = dx+model.q;
   H  = gf+model.q;
   
   f=ukr_g_errgradhess(gf,H,x,model,W,gs,0);
   lambda = 0.1*f;
   
   for (iter=0;iter<MAX_ITER;iter++) {   
      int i;
          
      ngf=vectorSum2(gf,model.q);
      if (ngf < GRAD_TOL) return f;
      
      for (i=0;i<model.q;i++) {
         H[i + i*model.q]+=lambda; 
      }
      invertSymmetricMatrix(H,model.q);
      multMatrixVector(dx,H,gf,model.q,model.q);

      if (vectorSum2(dx,model.q)<MIN_STEP) return f;
      
      pred = -0.5*dotProduct(gf,dx,model.q);
      for (i=0;i<model.q;i++) {
         x[i]-=dx[i];
      }
      
      oldf = f;
      f=ukr_g_errgradhess(gf,H,x,model,W,gs,0);
        
      if (f>oldf + 0.5*pred) {
         lambda*=5;
      } else {
         lambda*=0.9;
      }
   }
   return f;
}

void ukr_aux_clip_x(double *x,const double *lb,const double *ub,int q) {
   int i;
   for (i=0;i<q;i++) {
      if (x[i]<lb[i]) {
         x[i]=lb[i];
      } else if (x[i]>ub[i]) {
         x[i]=ub[i];
      }
   }
}

int ukr_aux_handle_boxcon(int *nb,double *g,double *H,const double *x,const double *lb,const double *ub,int q) {
   int i;
   int inb=0;   
   
   for (i=0;i<q;i++) {
      if (x[i]==lb[i] && g[i]>0) continue;
      if (x[i]==ub[i] && g[i]<0) continue;
      nb[inb++]=i;
   }
   /* delete rows and columns of H with binding constraints */
	if (inb>0 && inb<q) {
      int j;
   	for (i=0;i<inb;i++) {
			for (j=0;j<inb;j++) {
				H[j+i*inb]=H[nb[j]+nb[i]*q];
			}
			g[i] = g[nb[i]];
		}
	}
   return inb;
}


double ukr_g_opt_boxcon(double *x, UKR_Model model, UKR_Workspace *W, UKR_G_Struct gs) {
   int iter;
   double oldf,f,ngf,pred,lambda;
   double *H,*gf,*dx;
   int *nb;
   
   if (model.lb==NULL) return ukr_g_opt_unc(x,model,W,gs);

   dx = W->extraG;
   gf = dx+model.q;
   H  = gf+model.q;
   nb = W->nonbind;
   
   ukr_aux_clip_x(x,model.lb,model.ub,model.q);
   f=ukr_g_errgradhess(gf,H,x,model,W,gs,0);
   lambda = 0.1*f;
   
   for (iter=0;iter<MAX_ITER;iter++) {   
      int i;
      int nnb;
      
      nnb = ukr_aux_handle_boxcon(nb,gf,H,x,model.lb,model.ub,model.q);
      
      if (nnb == 0) return f;
      /*      
      for (i=0;i<nnb;i++) {
         int j;
         for (j=0;j<nnb;j++) {
            printf("%8.4f ",H[i+j*nnb]);
         }
         printf("\n");
      }
      */
          
      ngf=vectorSum2(gf,nnb);
      if (ngf < GRAD_TOL) return f;
      
      for (i=0;i<nnb;i++) {
         H[i + i*nnb]+=lambda; 
      }
      invertSymmetricMatrix(H,nnb);
      multMatrixVector(dx,H,gf,nnb,nnb);

      if (vectorSum2(dx,nnb)<MIN_STEP) return f;
      
      pred = -0.5*dotProduct(gf,dx,nnb);
      for (i=0;i<nnb;i++) {
         x[nb[i]]-=dx[i];
      }
      
      oldf = f;
      ukr_aux_clip_x(x,model.lb,model.ub,model.q);      
      f=ukr_g_errgradhess(gf,H,x,model,W,gs,0);
        
      if (f>oldf + 0.5*pred) {
         lambda*=5;
      } else {
         lambda*=0.9;
      }
   }
   return f;
}


void ukr_aux_pdqp_slc(double *x,const double *invA,const double *a,const double *b,double b2,double b0,double *tmp,int q) {
/* Find minimum x of x'Ax + a'x  s.t. b'x<=-b0 */
   double dpxb;
   double lambda;
   int i;
   
   multMatrixVector(x,invA,a,q,q);
   dpxb=dotProduct(x,b,q);
   
   if (dpxb >= b0) return;
   
   multMatrixVector(tmp,invA,b,q,q); /* Ab in Matlab code */
   lambda = dpxb/dotProduct(b,tmp,q);
   
   for (i=0;i<q;i++) {
      x[i] += b[i]*(b0/b2) - lambda*tmp[i];
   } 
}


double ukr_g_opt_denscon(double *x, UKR_Model model, UKR_Workspace *W, UKR_G_Struct gs) {
   int iter;
   double c,oldf,f,ngf,ngc,pred,lambda;
   double *H,*gf,*gc,*dx,*oldx;
   
   if (model.eta<=0) return ukr_g_opt_unc(x,model,W,gs);

   dx = W->extraG;
   oldx = dx+model.q;
   gf = oldx+model.q;
   gc = gf+model.q;
   H  = gc+model.q;
   
   c = ukr_g_congrad(gc,x,model,W);
   if (c>0) return -1; /* invalid init. */
   ngc=vectorSum2(gc,model.q);   
   
   f=ukr_g_errgradhess(gf,H,x,model,W,gs,1);
   lambda = 0.1*f;
   
   for (iter=0;iter<MAX_ITER;iter++) {   
      int i,iterC;
      
      ngf=vectorSum2(gf,model.q);

      if (ngf < GRAD_TOL) return f;
      
      if (c>CON_SENS && ngc>0) {
         double measure = -dotProduct(gc,gf,model.q)/sqrt(ngf*ngc);
         if (measure > KKT_SENS) return f;
      }
      
      for (i=0;i<model.q;i++) {
         H[i + i*model.q]+=lambda; 
      }
      invertSymmetricMatrix(H,model.q);
      ukr_aux_pdqp_slc(dx,H,gf,gc,ngc,c,oldx,model.q);

      if (vectorSum2(dx,model.q)<MIN_STEP) return f;
      
      pred = -0.5*dotProduct(gf,dx,model.q);
      for (i=0;i<model.q;i++) {
         oldx[i]=x[i];
         x[i]-=dx[i];
      }
      
      c = ukr_g_congrad(gc,x,model,W);
      ngc = vectorSum2(gc,model.q);
      
      iterC=0;
      while (c>0) {
         if (ngc==0) {
            for (i=0;i<model.q;i++) x[i]=0.5*(x[i]+oldx[i]);
         } else {
            /* step a little bit too far within the linear approximation */
            double step = -(c+1e-8)/ngc;
            for (i=0;i<model.q;i++) x[i]+=step*gc[i];
         }
         if (++iterC > 50 || sqDist(x,oldx,model.q)<MIN_STEP) {
            for (i=0;i<model.q;i++) x[i]=oldx[i];
            return f;
         }
         c = ukr_g_congrad(gc,x,model,W);
         ngc = vectorSum2(gc,model.q);
      }
      
      oldf = f;
      f=ukr_g_errgradhess(gf,H,x,model,W,gs,0);
        
      if (f>oldf + 0.5*pred) {
         lambda*=5;
      } else {
         lambda*=0.9;
      }
   }
   return f;
}
