/*! \file ukr.h
   \brief UKR Toolbox header file
   
   This C library provides some low level functions of the UKR algorithm.
   It mainly servers as the basis for the MEX-replacements of sume functions
   within the Matlab UKR toolbox, but it can be used without Matlab.
   
   For easy inter-operatability, the data format of full and sparse
   vectors and matrices used in this library is that of Matlab (or Fortran).
   That is, vectors are just arrays of doubles. Matrices are also 1-D arrays
   of doubles, with the elements stored in column-major order. 
   A 2x2 matrix 
   \f[M = \left(\begin{array}{cc}
   m_{11} & m_{12} \\
   m_{21} & m_{22}
   \end{array}\right)\f]
   is thus stored as\code
         double M[4] = {m11,m21,m12,m22};
   \endcode

*/

#ifndef __UKR_H
#define __UKR_H

/** Enumeration of density kernels that the UKR toolbox provides */
typedef enum {
   UKR_KERNEL_UNKNOWN=0,      /**< unknown kernel */
   UKR_KERNEL_GAUSSIAN=1,     /**< Gaussian kernel */
   UKR_KERNEL_SPARSE=100,     /**< Placeholder only, kernels with finite support get numbers > 100 */
   UKR_KERNEL_QUARTIC=102,    /**< Quartic kernel, square of Epanechnikov */
   UKR_KERNEL_TRIWEIGHT=103   /**< Triweight kernel (Epanechnikov kernel raised to power 3) */
} UKR_Kernel;

/** Enumeration of loss functions that the UKR toolbox provides */
typedef enum {
   UKR_LOSS_UNKNOWN=-1,    /**< unknown loss function */
   UKR_LOSS_L2=0,          /**< standard squared Euclidean loss function */
   UKR_LOSS_HUBER=1,       /**< Hubers robust loss function */
   UKR_LOSS_EPS2=2,        /**< Rotation symmetric epsilon-insensitive loss function */
   UKR_LOSS_EPS=3          /**< epsilon-insensitive loss function */
} UKR_Loss;

/** Main data structure for describing a UKR model */
typedef struct {
   const double *X;        /**< Matrix of latent variables (support) */
   const double *Xt;       /**< Matrix of latent variables (target) */
   const double *Y;        /**< Matrix of data vectors (support) */
   const double *Yt;       /**< Matrix of data vectors (target) */
   const double *G;        /**< Gramian or kernel matrix (Feature space UKR) */
   const double *mmm;      /**< Multimap metric, used for PSOM-like multimap projections */
   int q;                  /**< Latent space dimensionality */
   int d;                  /**< Data space dimensionality */
   int N;                  /**< Number of data vectors and latent variables, size of the model (support) */
   int Nt;                 /**< Number of data vectors and latent variables, size of the model (target) */
   int degree;             /**< Degree of local polynomials (0..3) */   
   const int *MaskIs;      /**< Row indices of elements to be excluded for cross-validation, in Matlab-compatible form */
   const int *MaskJs;      /**< Columns indices of elements to be excluded for cross-validation, in Matlab-compatible form */
   const double *epsilon;  /**< Parameter(s) for the loss function, not needed for standard \a UKR_LOSS_L2 */
   int epsM;               /**< Number of rows of \a epsilon (usually 1 or d) */
   int epsN;               /**< Number of columns of \a epsilon (usually 1 or N) */
   UKR_Kernel kernel;      /**< Density kernel, \see UKR_Kernel */
   UKR_Loss lossFunction;  /**< Loss function, \see UKR_Loss */
   double eta;             /**< Density threshold */
   const double *lb;       /**< Box constraints, lower bound */
   const double *ub;       /**< Box constraints, upper bound */   
} UKR_Model;

/** Data structure used for projecting new data onto the manifold (ukr_g) */
typedef struct {
   const double *y;        /**< Data vector to project */
   const double *KyY;      /**< Feature space dot products between new data and the models data */
   double Kyy;             /**< Feature space norm of the new data */
} UKR_G_Struct;

/** Data structure used for calculating the reconstruction error and gradient (ukr_errgrad) */
typedef struct {
   int *ir;
   int nz;
   double *K;
   double sumK;
   double *P;
   double *B;
   double *M;
   double *MK;
   double *C;
   double *tmp_d;
   double *Psi;
   double *GammaPsi;
   double *XQ_T;
   double *Xx;
   double *extra;
   double *extraG;
   int *nonbind;
} UKR_Workspace;

typedef struct {
   double penalty;
   double minDens;
   int numViolate;
} UKR_DensPenalty;

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*! Ukr_kp_dists_sparse is the general type of functions which
    calculate vectors of kernel functions, outer derivatives
    and the sum of basis functions from a vector of distances
    for a sparse density kernel.
    \see ukr_kp_dists_quartic
    \see ukr_kp_dists_triweight
*/
typedef double (*Ukr_kp_dists_sparse)(int,const double *,double *,double *,int *,int *,int,const int*);

/*! Ukr_kp_dists_full is the general type of functions which
    calculate vectors of kernel functions, outer derivatives
    and the sum of basis functions from a vector of distances
    for a non-sparse density kernel.
    \see ukr_kp_dists_gaussian
*/
typedef double (*Ukr_kp_dists_full)(int,const double *,double *,double *,int,const int*);

/*! ukr_get_kernel_sparse returns the right latent kernel 
    function. This is used internally throughout the toolbox.
    Returns NULL for unsupported kernel types, or if the desired
    kernel is not sparse.
    \see ukr_kp_dists_quartic, ukr_kp_dists_triweight
*/
Ukr_kp_dists_sparse ukr_get_kernel_sparse(UKR_Kernel kernel);

/*! ukr_get_kernel_full returns the right latent kernel 
    function. This is used internally throughout the toolbox.
    Returns NULL for unsupported kernel types, or if the desired
    kernel is sparse.
*/
Ukr_kp_dists_full ukr_get_kernel_full(UKR_Kernel kernel);

void ukr_kp_dists(const double *dist, UKR_Workspace *W, UKR_Model model, int numExclude, const int *exclude);

/*! Evaluates the UKR regression function f(x;X) for given a argument \a x 
    \return density estimate p(x) in latent space, or -1.0 if an error occured                         
    \param[out] f resulting function value (must point to an array of at least d doubles)
    \param[in] x position in latent space
    \param[in] model UKR model, \see UKR_Model
*/
double ukr_f(double *f,const double *x,UKR_Model model,UKR_Workspace *W);

/*! Evaluates the UKR regression function f(x;X) and its Jacobian matrix
    for given a argument \a x 
    \return density estimate p(x) in latent space, or -1.0 if an error occured                         
    \param[out] f resulting function value (must point to an array of at least d doubles)
    \param[out] J resulting Jacobian matrix (must point to an array of at least d*q doubles)
    \param[in] x position in latent space
    \param[in] model UKR model, \see UKR_Model
*/
double ukr_fJ(double *f,double *J,const double *x,UKR_Model model, UKR_Workspace *W);

/*! Evaluates the Gaussian kernel as a function of the squared euclidean distance
   \param[in] M length of the vector \a dist
   \param[in] dist vector containing the distances
   \param[out] b resulting vector of basis functions (Gaussian kernel), must 
               point to an array of at least \a M doubles
   \param[out] p vector of outer derivatives, may be NULL, or must point to
               an array of at least \a M doubles
   \param[in]  numExclude the number of elements to exclude from evaluation \n
               For normal usage, pass \a numExclude=0
   \param[in]  exclude the elements to exclude. \n
               This is useful for Leave-One-Out or Leave-K-Out cross-validation. \n
               Elements are excluded by setting their entry in \a b and \a p to 0.                  
   \return     Sum of basis functions (=|\a b|), or -1.0 if an error occured                                        
*/
double ukr_kp_dists_gaussian(int M,const double *dist,double *k,double *p,int numExclude,const int *exclude);

/*! Evaluates the Quartic kernel as a function of the squared euclidean distance
   \param[in] M the length of the vector \a dist
   \param[in] dist the vector containing the distances
   \param[out] k resulting vector of kernel functions (Quartic kernel)
               in sparse storage, must point to an array of at least \a M doubles
   \param[out] p vector of outer derivatives in sparse storage. 
               May be NULL, or must point to an array of at least \a M doubles
   \param[out] ir vector of indexes of non-zero basis functions, must point
               to an array of at least \a M doubles
   \param[out] nz number of non-zero basis functions
   \param[in]  numExclude the number of elements to exclude from evaluation \n
               For normal usage, pass \a numExclude=0
   \param[in]  exclude the elements to exclude. \n
               This is useful for Leave-One-Out or Leave-K-Out cross-validation. \n
               Elements are excluded by setting their entry in \a b and \a p to 0.                  
   \param[in]  normalize This flag determines if b should be normalized to |\a b|=1 \n
               If normalization is desired (\a normalize!=0) \ap gets scaled as well
   \return     Sum of kernel functions (=|\a b|), or -1.0 if an error occured                         
*/
double ukr_kp_dists_quartic(int M,const double *dist,double *k,double *p,int *ir,int *nz,int numExclude,const int *exclude);

/*! Evaluates the Quartic kernel as a function of the squared euclidean distance
    \see ukr_kp_dists_gaussian
*/    
double ukr_kp_dists_triweight(int M,const double *dist,double *b,double *p,int *ir,int *nz,int numExclude,const int *exclude);

/*! Evaluates the loss function as defined by \a model.loss 
      \param[in] col Current column, used only to select the right loss function parameter \a model.epsilon        
      \param[in,out] yr When called, the data vector as reconstructed by the UKR function\n
                        At return, the gradient of the loss function w.r.t. \a yr
      \param[in] y Original data vector                        
      \param[in] model UKR model
      \return Loss function L(\a yr-\a y), e.g. the squared euclidean distance between \a y and \a yr
*/            
double ukr_loss_function(int col, double *yr,const double *y,UKR_Model model);

/*! Select an UKR kernel function by its name, used for Matlab inter-operation
      \param[in] name Kernel function name ("Gaussian", "Quartic", "Triweight")
      \return UKR kernel as enumerated in ::UKR_Kernel
*/      
UKR_Kernel ukr_kernel_by_name(const char *name);

/*! Select an UKR loss function by its name, used for Matlab inter-operation
      \param[in] name Loss function name ("L2", "Huber", ...)
      \return UKR loss as enumerated in ::UKR_Loss
*/      
UKR_Loss ukr_loss_by_name(const char *name);

/*! Evaluate the UKR reconstruction error and (optionally) its gradient
    \b The gradient is added to \a grad !!!
      \param[out] grad Gradient w.r.t. \a model.X, must be NULL or point to an array of at least q*N doubles, to which the gradient is ADDED
      \param[in] model UKR model, including the current latent variables model.X\a
            A pre-calculated neighbor mask can be passed in model.MaskIs and model.MaskJs,
            in which case leave-K-out cross-validation is used
      \param[in] forceCV If this flag is non-zero, leave-one-out cross-validation is used,
            regardless of any values in \a model.MaskIs and \a model.MaskJs      
            If you want no cross-validation, pass \a forceCV=0 and \a model.MaskIs=model.MaskJs=NULL.            
      \return UKR reconstruction error, or -1.0 if an error occured                   
*/      
double ukr_errgrad(double *GX, double *GXt,UKR_Model model,UKR_Workspace *W,int forceCV);

/*! Evaluate the UKR reconstruction error and (optionally) its gradient\a
      
      \param[out] grad Gradient w.r.t. \scale, must be NULL or point to an array of at least q doubles
      \param[in] scale Scaling factor(s) (inverse bandwidth)
      \param[in] model UKR model, including the current latent variables model.X\a
            A pre-calculated neighbor mask can be passed in model.MaskIs and model.MaskJs,
            in which case leave-K-out cross-validation is used
      \param[in] forceCV If this flag is non-zero, leave-one-out cross-validation is used,
            regardless of any values in \a model.MaskIs and \a model.MaskJs      
            If you want no cross-validation, pass \a forceCV=0 and \a model.MaskIs=model.MaskJs=NULL.            
      \return UKR reconstruction error, or -1.0 if an error occured          
      \see ukr_errgrad
*/      
double ukr_errgrad_scale(double *grad, const double *s,UKR_Model model,UKR_Workspace *W,int forceCV);

void ukr_aux_shiftX(double *Xx, int q, int nz, const int *ir);

int ukr_aux_b(UKR_Workspace *W, UKR_Model model);
int ukr_aux_bJ(double *J,UKR_Workspace *W,UKR_Model model);
double ukr_aux_M(int col,UKR_Workspace *W,UKR_Model model);

void ukr_aux_l_b(double *b,const double *Xx,const double *invC,const double *K,int q,int N);
void ukr_aux_q_b(double *b,const double *Xx,const double *invC,const double *K,int q,int N);
void ukr_aux_q2_b(double *b,const double *Xx,const double *invC,const double *K,int N);

void ukr_aux_l_C(double *C,const double *Xx,const double *K,int q,int N);
void ukr_aux_q_C(double *C,const double *Xx,const double *K,int q,int N);
void ukr_aux_q2_C(double *C,const double *Xx,const double *K,int N);

void multXiTMatrix_l(double *D,const double *Xx,const double *M,int q,int N,int d);
void multXiTMatrix_q(double *D,const double *Xx,const double *M,int q,int N,int d);
void multXiTMatrix_q2(double *D,const double *Xx,const double *M,int N,int d);
void multXiMatrix_l(double *D, const double *Xx, const double *M, int q,int N,int d);
void multXiMatrix_q(double *D, const double *Xx, const double *M, int q,int N,int d);
void multXiMatrix_q2(double *D, const double *Xx, const double *M, int N,int d);
void ukr_aux_XQ_T(UKR_Workspace *W, int degree, int q, int Nt);
 
UKR_DensPenalty ukr_dens_penalty(double *GX, double *GXt,UKR_Model model, UKR_Workspace *W);
double ukr_densvar_penalty(double *GX,double *GXt,UKR_Model model,UKR_Workspace *W);

int ukr_aux_alloc_workspace(UKR_Model model, UKR_Workspace *W);
void ukr_aux_free_workspace(UKR_Workspace W);

double ukr_g_congrad(double *g,const double *x,UKR_Model model, UKR_Workspace *W);
double ukr_g_errgradhess(double *g,double *H,const double *x,UKR_Model model, UKR_Workspace *W, UKR_G_Struct ukr_gs,int bp_ready);

double ukr_g_opt_unc(double *x, UKR_Model model, UKR_Workspace *W, UKR_G_Struct gs);
double ukr_g_opt_boxcon(double *x, UKR_Model model, UKR_Workspace *W, UKR_G_Struct gs);
double ukr_g_opt_denscon(double *x, UKR_Model model, UKR_Workspace *W, UKR_G_Struct gs);

void ukr_aux_clip_x(double *x,const double *lb,const double *ub,int q);
int ukr_aux_handle_boxcon(int *nb,double *g,double *H,const double *x,const double *lb,const double *ub,int q);
void ukr_aux_pdqp_slc(double *x,const double *invA,const double *a,const double *b,double b2,double b0,double *tmp,int q);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
