#include <mex.h>
#include <matrix.h>
#include "ukr.h"
#include "ukr_matlab.h"

UKR_Model ukr_model_from_matlab(const mxArray *MS) {
   UKR_Model model;
   mxArray *ar;
   char str[20];
   
   model.q=model.d=model.N=model.Nt=0;
   model.X=model.Y=model.G=model.Yt=model.mmm=NULL;
   model.MaskIs=model.MaskJs=NULL;
   model.lb=model.ub=NULL;
   
   ar = mxGetField(MS,0,"q");
   if (ar) {
      model.q=(int) mxGetScalar(ar);
   }
   
   ar = mxGetField(MS,0,"X");
   if (ar) {
      int q=mxGetM(ar);
      if (model.q==0) {
         model.q=q;
      } else {
         if (model.q!=q) mexErrMsgTxt("Dimension of model.X does not match model.q");
      }
      model.N=mxGetN(ar);
      model.X=mxGetPr(ar);
   }
   
   ar = mxGetField(MS,0,"Xt");
   if (ar) {
      int q=mxGetM(ar);
      if (model.X==NULL) mexErrMsgTxt("You cannot pass model.Xt without passing model.X");
      if (model.q!=q) mexErrMsgTxt("Dimension of model.Xt does not match model.q or model.X");
      model.Nt=mxGetN(ar);
      model.Xt=mxGetPr(ar);
   } 
   
   ar=mxGetField(MS,0,"G");
   if (ar) {
      int N=mxGetM(ar);
      if (model.N==0) {
         model.N=N;
      } else {
         if (model.N!=N) mexErrMsgTxt("Dimension of model.G does not match model.X");
      }
      if (mxGetN(ar)!=N) mexErrMsgTxt("model.G must be an NxN matrix.");
      model.G=mxGetPr(ar);
   }
         
   if (model.G==NULL) {   
      ar=mxGetField(MS,0,"Y");
      if (ar!=NULL) {
         int N=mxGetN(ar);      
         if (model.N==0) {
            model.N=N;
         } else {
            if (model.N!=N) mexErrMsgTxt("Dimension of model.Y does not match model.X");
         }
         model.d=mxGetM(ar);
         model.Y=mxGetPr(ar);
      }

      ar=mxGetField(MS,0,"Yt");
      if (ar) {
         int d,Nt;
         d=mxGetM(ar);
         Nt=mxGetN(ar);
         if (model.Y==NULL) mexErrMsgTxt("You cannot pass model.Yt without model.Y");
         if (model.d!=d) mexErrMsgTxt("Dimension of model.Yt does not match model.Y");
         if (model.Nt==0) {
            model.Nt=Nt;
         } else {
            if (Nt!=model.Nt) mexErrMsgTxt("Dimension of model.Yt does not match model.Xt");
         }
         model.Yt=mxGetPr(ar);
      } else {
         model.Yt=model.Y;
      }
   }
   
   if (model.Nt==0) {
      model.Nt=model.N;
   }
   
   /* By now, we have 3 cases:
      a) None of X,Xt,Y,Yt or G has been given -> N=Nt=0
      b) One of X,Y,G has been given -> N=Nt is set
      c) (Y,Yt) or (X,Xt) have been given -> N and Nt are set, might be different
   */
   
   ar=mxGetField(MS,0,"kernel_l");
   if (ar==NULL) mexErrMsgTxt("No latent kernel function given.");
   mxGetString(ar,str,20);
   model.kernel=ukr_kernel_by_name(str);
   if (model.kernel==UKR_KERNEL_UNKNOWN) mexErrMsgTxt("Unknown latent kernel function.");
   
   ar=mxGetField(MS,0,"eta");
   if (ar==NULL) {
      model.eta=0;
   } else {
      model.eta=mxGetScalar(ar);
   }  
   
   ar=mxGetField(MS,0,"lb");
   if (ar!=NULL) {
      int bq = mxGetM(ar);
      int i;
      
      if (model.q==0) model.q=bq;
      if (model.q!=bq || mxGetN(ar)!=1) mexErrMsgTxt("Lower bound (model.lb) must be a 'q x 1' vector.");
      model.lb = mxGetPr(ar);
      
      ar=mxGetField(MS,0,"ub");
      if (ar==NULL) {
         mexErrMsgTxt("You may not give a lower bound (model.lb) without an upper bound (model.ub).");
      }
      model.ub = ukr_mx_get_matrix(ar,bq,1);
      if (model.ub==NULL) mexErrMsgTxt("Upper bound (model.ub) must be a 'q x 1' vector.");
      
      for (i=0;i<bq;i++) {
         if (model.ub[i]<=model.lb[i]) mexErrMsgTxt("Given lower and upper bounds are not valid.");     
      }
   }
   
   ar=mxGetField(MS,0,"degree");
   if (ar==NULL) {
      model.degree=0;
   } else {
      model.degree=(int) mxGetScalar(ar);
      switch(model.degree) {
         case 0:
         case 1:
         case 2:
            break;
         case 3:
           if (model.q!=2) mexErrMsgTxt("Full quadratic model (degree=3) is only available for model.q=2");
           break;
         default:
           mexErrMsgTxt("model.degree must be between 0 and 3.");
           return;
      }
   }  
   
   ar=mxGetField(MS,0,"Mask");
   if (ar!=NULL) {
      int N,Nt;
      if (!mxIsSparse(ar)) mexErrMsgTxt("model.Mask (if given) must be sparse!");
      
      N=mxGetM(ar);
      Nt=mxGetN(ar);
      
      if (model.N==0) {
         model.N=N;
         model.Nt=Nt;
      } else {
         if (model.N!=N || model.Nt!=Nt)
            mexErrMsgTxt("Dimensions of model.Mask do not match rest of model");
      }  
      model.MaskIs=mxGetIr(ar);
      model.MaskJs=mxGetJc(ar);
   }
   
   /* If no data matrix is given, ignore the loss function and the multimap metric */
   if (model.Y==NULL) return model;
      
   ar=mxGetField(MS,0,"loss");
   if (ar==NULL) {
      model.lossFunction=UKR_LOSS_L2;
   } else {
      mxGetString(ar,str,20);   
      model.lossFunction=ukr_loss_by_name(str);
   }
   
   if (model.lossFunction!=UKR_LOSS_L2) { /* loss functions that need parameters */
      int validEps;
      
      ar=mxGetField(MS,0,"eps");
      if (ar==NULL) mexErrMsgTxt("Given loss function requires a parameter [model.eps].");
      
      model.epsilon=mxGetPr(ar);               
      model.epsM=mxGetM(ar);
      model.epsN=mxGetN(ar);
      
      switch(model.lossFunction) {
         case UKR_LOSS_EPS:
            if (model.epsN==1) {
               validEps = (model.epsM==1) || (model.epsM==model.d);
            } else if (model.epsN==model.N) {
               validEps = (model.epsM==model.d);
            } else {
               validEps = 0;
            }
            break;
         
         case UKR_LOSS_EPS2:
            validEps = (model.epsM==1) && (model.epsN==1 || model.epsN==model.N);
            break;
         
         case UKR_LOSS_HUBER:
            validEps = (model.epsN==1) && (model.epsM==1 || model.epsM==model.d);
            break;
         default:
            mexErrMsgTxt("Unknown loss function.");
            break;         
      }
      if (!validEps) mexErrMsgTxt("size(model.eps) does not match with choice of loss function.");
   }
   
   ar=mxGetField(MS,0,"mmm");
   if (ar!=NULL) {
      if (mxGetM(ar)!=model.d || mxGetN(ar)!=1) {
         mexErrMsgTxt("model.mmm (if given) must be a column vector matching the data space dimensionality.");
      }
      model.mmm = mxGetPr(ar);
   }
   
   return model;
}

const double *ukr_mx_get_matrix(const mxArray *ar, int m,int n) {
   if (ar==NULL) return NULL;
   if (!mxIsDouble(ar)) return NULL;
   if (mxGetM(ar)!=m) return NULL;
   if (mxGetN(ar)!=n) return NULL;
   return mxGetPr(ar);
}


UKR_G_Struct ukr_g_struct_from_matlab(const mxArray *ar, int d, int N) {
   UKR_G_Struct gs;
   gs.y = gs.KyY = NULL;
   
   if (mxIsDouble(ar) && mxGetM(ar)==d && mxGetN(ar)==1) {
      gs.y = mxGetPr(ar);
      return gs;
   } else if (mxIsStruct(ar)) {
      const mxArray *yD;
      const mxArray *yd;
      
      yD = mxGetField(ar,0,"D");
      yd = mxGetField(ar,0,"d");
      if (yD!=NULL && mxGetM(yD)==N && mxGetN(yD)==1 &&
          yd!=NULL && mxGetM(yd)==1 && mxGetN(yd)==1) {
         gs.KyY = mxGetPr(yD);
         gs.Kyy = mxGetScalar(yd);
         return gs;
      } 
   } else {
      mexErrMsgTxt("Parameter y must be either a column vector (data space) or a struct as returned by ukr_mercer (feature space).");
   }
   return gs;
}
