function [err, grad] = ukr_errgrad_ol(X, model, nex, cv)
% [err, grad] = ukr_errgrad(X, model, nex, [,cv]) 
%
% Compute mean squared reconstruction error and its gradient
%
% The flag "cv" determines if cross-validation is used
% If not given, model.Mask determines the kind of cross-validation
%   cv=0 : do not use cross-validation (overrides model.Mask)
%   cv~=0: use leave-one-out cross-validation (overrides model.Mask)
%
% Normally you don't call this function directly.

error(nargchk(3, 4, nargin))

[q,Na]=size(X);

if isfield(model,'G')
   error 'Feature space UKR not (yet) supported for online training';
end

[d,N]=size(model.Y);

if isfield(model,'Yt')
   Nt = size(model.Yt,2);
   T  = model.Yt(:,nex);
   Xt = X(:,N+nex);
   plain=0;
else
   Nt = N;
   T = model.Y(:,nex);
   Xt = X(:,nex);
   plain=1;
end

if nargin>2
   if cv
      if N~=Nt
         error 'LOO-CV cannot be applied in this case.'
      end
      Mask=sparse(nex,1,1,N,1,1);
   else
      Mask = sparse([],[],[],N,1,0);
   end
else
   if isfield(model,'Mask')
      if ~isequal(size(Mask),[N Nt])
         error 'Given CV-Mask does not match given latent variable matrix';
      end
      Mask = model.Mask(:,nex);
   else
      Mask = sparse([],[],[],N,1,0);
   end
end

if plain 
   [err,grad,GXt] = ukr_aux_errgrad(X,Xt,T,model,Mask);
   grad(:,nex) = grad(:,nex) + GXt;
else
   [err,GX,GXt] = ukr_aux_errgrad(X(:,1:N),Xt,T,model,Mask);
   grad = [GX zeros(q,nex-1) GXt zeros(q,Nt-nex)];
end
