function [X] = ukr_g(Y, model)
%UKR_G Evaluate ukr backward mapping.
% [X] = ukr_g(Y, model)
%     or
% [X] = ukr_g(T, model)
%
% Inputs:
%   Y      -  Observable data space elements (in columns).
%
%   model  -  Valid ukr_model structure.
%
%   T      -  Structure as returned by ukr_mercer :
%   T.D    -  Matrix of dot products in feature space
%   T.d    -  Row vector of dot products ...
%             This should only be used if the observables are not explictly
%             given, or you did not set "model.kernel_o" to a valid
%             function handle (mercer kernel in observable space)
%
% Output:
%   X      -  Result of applying the ukr-'backward mapping' as implicitly
%             defined by the input ukr model to input Y.
%
% Yields the 'projection index' of the orthogonal projection
% of input y onto the (density-constrained) ukr manifold.

error(nargchk(2, 2, nargin))

kernelMode=0;

if ~isstruct(Y)
   [d1 M]=size(Y);
   [d2 N]=size(model.Y);

   if (d1~=d2) error('Dimensions (training/test data) do not match'); end;

   if isfield(model,'kernel_o')
      kernelMode=1;
      Y=ukr_mercer(model,Y);
   end
else
   kernelMode=1;
   [N M]=size(Y.D);
   if (N~=size(model.G,1)) | (M~=size(Y.d,2))
      error('Dimensions of dot product matrices are not correct');
   end
end

if ~kernelMode
   % Initialize with 'nearest reconstructions':
   [Ynr index] = ukr_nr(Y, model);
   X=model.X(:,index); % initial values for optimization
   if isfield(model,'eta') & model.eta > 0
      for k=1:M
         X(:,k)=ukr_g_opt_denscon(Y(:,k),model,X(:,k));
      end
   elseif isfield(model,'lb')
      for k=1:M
         X(:,k)=ukr_g_opt_boxcon(Y(:,k),model,X(:,k));
      end
   else
      for k=1:M
         X(:,k)=ukr_g_opt_unc(Y(:,k),model,X(:,k));
      end
   end
else
   index = ukr_nfr(Y,model);
   X=model.X(:,index); % initial values for optimization
   if isfield(model,'eta') & model.eta > 0
      for k=1:M
         T.D=Y.D(:,k);
         T.d=Y.d(k);
         X(:,k)=ukr_g_opt_denscon(T,model,X(:,k));
       end
   elseif isfield(model,'lb')
      for k=1:M
         T.D=Y.D(:,k);
         T.d=Y.d(k);
         X(:,k)=ukr_g_opt_boxcon(T,model,X(:,k));
      end
   else
      for k=1:M
         T.D=Y.D(:,k);
         T.d=Y.d(k);
         X(:,k)=ukr_g_opt_unc(T,model,X(:,k));
      end
   end
end

%res=sum((ukr_f(X,model)-Y).^2);
%res2=sum((ukr_f(X2,model)-Y).^2);

%res-res2
%mean(res-res2)

