function [model] = ukr_model(varargin);
% UKR_MODEL Generate a valid ukr_model structure.
% [model] = ukr_model('field name',parameter,...)
% 
% Option-strings and corresponding inputs:
% (Note that the only mandatory argument is the observable space 
% dataset specified using the option-string 'Y'.)
%
%  'Y' (double-matrix)   
%        Observable space dataset Y.
%
%  'X' (double-matrix)   
%        Latent space dataset X.
%
%  'q' (double)          
%        Latent space dimensionality. If option 'q' is used, any 
%        user-defined initialization (using 'X') will be overridden.
%
%  'kernel_l' (string)
%        Latent space density kernel function. Currently supported 
%        functions are: 'Quartic','Triweight' and 'Gaussian'
%        The default is 'Quartic'.
%
%  'degree'   (double)
%        Degree of local polynomials.
%           0: Local constant fit, Nadaraya-Watson (default)
%           1: Local linear fit (seems to have no advantages)
%           2: Local quadratic fit w/o cross-terms
%           3: Local quadratic fit with cross-term for q=2
%
%  'kernel_o' (function handle) 
%        Observable space Mercer kernel function.
%
%  'G' (double matrix)   
%        Gramian of observable space dataset Y / Mercer kernel matrix 
%
%  'display' (double/logical)  
%        Flag : display optimization process 
%
%  'loss' (string)          
%        UKR loss function, at present either
%        'Eps', 'Eps2', 'L2' (default) or 'Huber'   
%
%  'eps'    (double matrix)
%        Loss function parameters, dimension must be 
%           'Eps'   -> 1x1, dx1 or dxN
%           'Eps2'  -> 1x1 or 1xN
%           'Huber' -> 1x1 or dx1      
%           'L2'    -> parameter is ignored 
%
%  'penalty' (function handle or string)  
%        Penalty function used for unconstrained optimization
%        Either  @ukr_ext_penalty, @ukr_dens_penalty or @ukr_densvar_penalty
%
%  'lambda' (double)
%        Pre-factor for penalty function (unconstrained opt.)
%
% Output: 
% model           -  Valid ukr_model structure.

model.q=1;
model.display=0;

X_set = logical(0); 
Y_set = logical(0); 
kernel_l_set = logical(0); 
eta_set = logical(0); 
K_set = logical(0); 
G_set = logical(0);
pen_set = 0;
loss_set = 0;
deg_set = 0;

ii=1; 
while ii<=length(varargin), 
  if ischar(varargin{ii}), 
    switch varargin{ii}, 
      case 'X';        
         ii=ii+1; 
         model.X = varargin{ii}; 
         X_set = logical(1); 
         model.q=size(model.X,1);
      case 'q';        
         ii=ii+1; 
         if X_set; 
            warning('Use of option \"q\" overrides user defined latent data matrix X!'); 
            X_set = logical(0); 
         end; 
         model.q = varargin{ii}; 
      case 'Y';        
         ii=ii+1; 
         model.Y = varargin{ii}; 
         Y_set = logical(1);
      case 'kernel_l'; 
         ii=ii+1; 
         model.kernel_l = varargin{ii}; 
         kernel_l_set = logical(1);
      case 'degree';
         ii=ii+1;
         model.degree = varargin{ii};
         deg_set = logical(1);
      case 'kernel_o'; 
         ii=ii+1; 
         model.kernel_o = varargin{ii}; 
         K_set=logical(1);
      case 'G';        
         ii=ii+1; 
         model.G = varargin{ii}; 
         G_set = logical(1);
      case 'eta';      
         ii=ii+1; 
         model.eta = varargin{ii}; 
         eta_set = logical(1);
      case 'display';  
         ii=ii+1; 
         model.display = varargin{ii}; 
      case 'penalty';
         ii=ii+1;
         model.penalty = varargin{ii};
         pen_set = pen_set+1;
      case 'lambda';
         ii=ii+1;
         model.lambda = varargin{ii};
         pen_set = pen_set+2;
      case 'loss';
         ii=ii+1;
         model.loss = varargin{ii};
         loss_set = loss_set+1;
      case 'eps';
         ii=ii+1;
         model.loss = varargin{ii};
         loss_set = loss_set+2;
      otherwise 
         error(['Invalid Identifier, (', varargin{ii}, ') !']); 
    end
  else
    error('Invalid Identifier.');
  end
  ii = ii+1; 
end

if ~Y_set
    if ~G_set
        error('Mandatory argument Y (or G) not given!'); 
    end
    N=size(model.G,1);
else
    [d N] = size(model.Y);
    if K_set & ~G_set
        model.G=ukr_mercer(model);
    end
end
   
if ~X_set; model.X = rand(model.q, N)-0.5; end;
if ~kernel_l_set; model.kernel_l = 'Quartic'; end;
if ~eta_set; model.eta = 0; end;
if pen_set == 1
   warning 'You tried to set \"penalty\" without setting \"lambda\"';
   model=rmfield(model,'penalty');
end
if pen_set == 2
   warning 'You tried to set \"lambda\" without setting \"penalty\"';
   model=rmfield(model,'lambda');
end
if loss_set == 1 & ~isequal(model.loss,'L2')
   warning 'You tried to set \"loss\" without setting \"eps\"';
   model=rmfield(model,'loss');
end
if loss_set == 2 
   warning 'You tried to set \"eps\" without setting \"loss\"';
   model=rmfield(model,'eps');
end
if loss_set == 3 & isequal(model.loss,'L2')
   warning '\"eps\" is ignored for the L2 loss function'
end
