function X = ukr_opt(Xinit, maxiter, eta_cons, model, cv)
% function X = ukr_opt(Xinit, maxiter, eta_cons, model [,cv])
%
% Wrapper for specialized optimization routines used during ukr_train.
%
%  Xinit    - initial latent variables
%  maxiter  - maximum number of RPROP iterations
%  eta_cons - minimal density thresholds for ...
%              a) homotopy based optimization
%                 Use a monotonous decreasing series of tresholds,
%                 the last being 0 if you want CV model selection
%                 (in this case, either use cv=1 or use model.Mask
%              b) density constraint optimization
%                 Use a non-zero value as the last (or single) entry 
%                 in eta_cons
%  model    - UKR model structure (see ukr_model)
%  cv       - pass 0 (no CV) or 1 (LOO-CV) to override model.Mask

X = Xinit;

if nargin == 5
   if cv == 1
      mode=2;
   else
      mode=0;
   end
else
   if isfield(model,'Mask')
      mode=1;
   else
      mode=0;
   end
end

if isfield(model,'penalty') & isfield(model,'lambda')
   mode=mode+10;
end

% mode=0: no CV, no penalty
% mode=1: LKO-CV (model.Mask)
% mode=2: LOO-CV (override model.Mask if existing)
% mode=10: penalty only
% mode=11: penalty + LKO-CV (model.Mask)
% mode=12: penalty + LOO-CV ((override model.Mask if existing)

if ~isempty(eta_cons) & ~isequal(eta_cons,0)
   if mode >= 10
      error 'You try regularizing with both a penalty term and a density constraint, which is not supported in this toolbox.';
   end

   for k=1:length(eta_cons)
      if eta_cons(k) == 0 
         switch(mode)
            case 0;
               warning('eta=0 without CV makes no sense, stopping.');
               return
            case 1;
               X = ukr_opt_unc(@ukr_errgrad,X,maxiter,model);
            case 2;
               X = ukr_opt_unc(@ukr_errgrad,X,maxiter,model,1);
         end
      else
         if mode==2
            X = ukr_opt_denscon(X,maxiter,model,eta_cons(k),cv);         
         else
            X = ukr_opt_denscon(X,maxiter,model,eta_cons(k));
         end
      end
   end
else
   switch(mode)
      case 0;
         error('No regularization method selected.');      
      case 1;
         X = ukr_opt_unc(@ukr_errgrad,X,maxiter,model);
      case 2;
         X = ukr_opt_unc(@ukr_errgrad,X,maxiter,model,1);
      case {10,11};
         X = ukr_opt_unc_penalty(@ukr_errgrad,model.penalty,...
               model.lambda,X,maxiter,model);      
      case 12;
         X = ukr_opt_unc_penalty(@ukr_errgrad,model.penalty,...
               model.lambda,X,maxiter,model,1);      
   end
end

