function [X, cand_ind] = ukr_train(model, candidates, steps, etas, loocv)
% UKR_TRAIN   Train a ukr model.
%
% X = ukr_train(model [, candidates [, steps [, eta [, loocv]]]])
%
% Inputs:
%
%   model        - valid ukr_model structure.
%
% candidates     - cell array of candidate initializations
%                  pass [] or {} if you only want (K)PCA and
%                  homotopy based training
%
%   steps        - number of RPROP steps during fine optimization
%                   (default: 500)
%
%   etas         - density constraint value(s), pass [] for pure CV
%
%   loocv        - logical switch to use leave-one-out cross-validation 
%                  or not, default is 1 (use LOO-CV)
%
% Output:
%
%   X            - latent space dataset resulting from training.
%
%   cand_ind     - index of selected candidate. If you passed K candidates,
%                  a value of (K+1) means the PCA initialization was selected
%
%
% Examples : 
%
%   for k=1:5; Xlle{k}=lle(model.Y,5+k,model.q); end
%   model.X=ukr_train(model,Xlle,200);
%
%      This first generates LLE solutions for 6..10 neighbours.
%      Additionally, a PCA solution is calculated. Then, optimal 
%      scales are found and the best candidate (LLE + PCA) is selected.
%      Finally, the best candidate is fine tuned with 200 RPROP steps
%      minimizing the CV-error.
%      If PCA yields the best candidate (rather seldom), homotopy 
%      based optimization is applied, using default density thresholds
%      etas = [0.2 0.1 0.05 0.02 0] with 200 steps each.
%
%   model.X=ukr_train(model,[],30,[0.1 0.05 0.02 0 0])
%
%       leads to 30 steps with constraint density>=0.1
%                30 steps with constraint density>=0.05
%                30 steps with constraint density>=0.02
%                60 unconstrained steps (cv only)
%             


error(nargchk(1,5,nargin))

if nargin<2
   candidates={};
end

if nargin<3
   steps=500;
end

if nargin<4
   eta_hom=[0.2 0.1 0.05 0.02 0 0]
   if nargin<3
      steps=100;
   end
   eta_c = 0;
else
   if length(etas)>1
      eta_hom = etas;
      eta_c = etas(end);
   else
      eta_hom = etas;
      eta_c = etas;
   end
end

if nargin<5
   loocv=1;
end

if eta_c==0 & ~loocv
   error 'Optimization using neither a density constraint nor leave-one-out crossvalidation normally leads to interpolation solutions';
end

if isfield(model,'Y')
   [d,N]=size(model.Y);
elseif isfield(model,'G')
   [N,N2]=size(model.G);
   if N~=N2 
      error 'model.G must be a symmetric NxN matrix.'
   end
else
   error 'Neither model.Y nor model.G given. Please see ukr_model.'
end   

if isfield(model,'q')
   q=model.q;
elseif isfield(model,'X')
   [q,N2]=size(model.X);
   model.q=q;
else
   error 'Please set model.q to the desired latent dimensionality.'
end

K=0;

if iscell(candidates)
   K=length(candidates);
   if K>0
      for k=1:K
         [qc,Nc]=size(candidates{k});
         if qc~=q | Nc~=N
            error 'Dimensions of candidate(s) do not match.'
         end
      end
      errs = zeros(K+1,1);
      S = zeros(q,K+1);
      for k=1:K
         [S(:,k), errs(k)]=ukr_initscale(candidates{k},model);
      end
   end
end      

Xpca = ukr_pca(model);

if K>0
   [S(:,K+1), errs(K+1)]=ukr_initscale(Xpca,model);
   [minerr,ind]=min(errs);
else
   ind=1;
end

if ind==K+1 % PCA is only or best
   fprintf(1,'\nUsing homotopy method...\n');
   X = ukr_opt(0.1*Xpca,steps,eta_hom,model,loocv);
else
   fprintf(1,'Using candidate %i\n',ind);
   if eta_c == 0 
      X = ukr_opt(diag(S(:,ind))*candidates{ind},steps,[],model,loocv);      
   else
      X = ukr_opt(diag(S(:,ind))*candidates{ind},steps,eta_c,model,loocv);
   end
end
  
if nargout>1
   cand_ind = ind;
end
